/************************************************************************
	rgbled.c

    USB RGB Bargraph
    Copyright (C) 2010 Simon Inns

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	Email: simon.inns@gmail.com

************************************************************************/

#ifndef RGBLED_C
#define RGBLED_C

// Global includes
#include <htc.h>

// Local includes
#include "hardware.h"
#include "rgbled.h"
#include "tlc5940.h"

// This procedure updates the RGB LED brightness levels according to the 
// current fading parameters
void procesRgbLedBrightness(void)	
{
	// Process the RGB LEDs
	for (unsigned char rgbLedNumber = 0; rgbLedNumber < 16; rgbLedNumber++)
	{
		// Process the red colour fading
		if (display.rgbLed[rgbLedNumber].redTarget != display.rgbLed[rgbLedNumber].redActual)
		{
			// If fade is off then just set the brightness
			if (display.rgbLed[rgbLedNumber].redFadeSpeed == -1)
				display.rgbLed[rgbLedNumber].redActual = display.rgbLed[rgbLedNumber].redTarget;
			else
			{
				// Fade the colour up or down
				if (display.rgbLed[rgbLedNumber].redTarget > display.rgbLed[rgbLedNumber].redActual)
					display.rgbLed[rgbLedNumber].redActual+= display.rgbLed[rgbLedNumber].redFadeSpeed;
				else display.rgbLed[rgbLedNumber].redActual-= display.rgbLed[rgbLedNumber].redFadeSpeed;
				
				// Range check the fade
				if (display.rgbLed[rgbLedNumber].redActual > 4095) display.rgbLed[rgbLedNumber].redActual = 4095;
				if (display.rgbLed[rgbLedNumber].redActual < 0) display.rgbLed[rgbLedNumber].redActual = 0;
			}
		}	
			
		// Process the green colour fading
		if (display.rgbLed[rgbLedNumber].greenTarget != display.rgbLed[rgbLedNumber].greenActual)
		{
			// If fade is off then just set the brightness
			if (display.rgbLed[rgbLedNumber].greenFadeSpeed == -1)
				display.rgbLed[rgbLedNumber].greenActual = display.rgbLed[rgbLedNumber].greenTarget;
			else
			{
				// Fade the colour up or down
				if (display.rgbLed[rgbLedNumber].greenTarget > display.rgbLed[rgbLedNumber].greenActual)
					display.rgbLed[rgbLedNumber].greenActual+= display.rgbLed[rgbLedNumber].greenFadeSpeed;
				else display.rgbLed[rgbLedNumber].greenActual-= display.rgbLed[rgbLedNumber].greenFadeSpeed;
				
				// Range check the fade
				if (display.rgbLed[rgbLedNumber].greenActual > 4095) display.rgbLed[rgbLedNumber].greenActual = 4095;
				if (display.rgbLed[rgbLedNumber].greenActual < 0) display.rgbLed[rgbLedNumber].greenActual = 0;
			}
		}	
		
		// Process the blue colour fading
		if (display.rgbLed[rgbLedNumber].blueTarget != display.rgbLed[rgbLedNumber].blueActual)
		{
			// If fade is off then just set the brightness
			if (display.rgbLed[rgbLedNumber].blueFadeSpeed == -1)
				display.rgbLed[rgbLedNumber].blueActual = display.rgbLed[rgbLedNumber].blueTarget;
			else
			{
				// Fade the colour up or down
				if (display.rgbLed[rgbLedNumber].blueTarget > display.rgbLed[rgbLedNumber].blueActual)
					display.rgbLed[rgbLedNumber].blueActual+= display.rgbLed[rgbLedNumber].blueFadeSpeed;
				else display.rgbLed[rgbLedNumber].blueActual-= display.rgbLed[rgbLedNumber].blueFadeSpeed;
				
				// Range check the fade
				if (display.rgbLed[rgbLedNumber].blueActual > 4095) display.rgbLed[rgbLedNumber].blueActual = 4095;
				if (display.rgbLed[rgbLedNumber].blueActual < 0) display.rgbLed[rgbLedNumber].blueActual = 0;
			}
		}		
			
		// Set the correct LED channels for the RGB LED
		setGrayScaleValue(3 * rgbLedNumber, display.rgbLed[rgbLedNumber].greenActual);
		setGrayScaleValue(3 * rgbLedNumber + 1, display.rgbLed[rgbLedNumber].blueActual);
		setGrayScaleValue(3 * rgbLedNumber + 2, display.rgbLed[rgbLedNumber].redActual);
	}
	
	// Update the TLC5940s
	while(updateTlc5940() == 1);
}

// Set an RGB value in the RGB LED state model and initialise the RGB fading parameters
void setRgbLedColour(unsigned char rgbLedNumber, int red, int green, int blue)
{	
	// Normalise the brightness output (stolen from my USB Performance meter project - see
	// there for details of what this scaling is for)
	float redScale = (47.0 * (float)red / 71.0);
	float greenScale = ((float)green);
	float blueScale = (47.0 * (float)blue / 64.0);
	
	red = (int)redScale;
	green = (int)greenScale;
	blue = (int)blueScale;
	
	// If the new values are the same as currently set, just return
	if ( (red == display.rgbLed[rgbLedNumber].redTarget) &&
		(blue == display.rgbLed[rgbLedNumber].blueTarget) &&
		(green == display.rgbLed[rgbLedNumber].greenTarget)) return;
	
	// Here we specify the minimum intensity (if we allow the intensity to be too low
	// there aren't enough brightness levels to fade smoothly
	if (red != 0 && red < 20) red = 20;
	if (green != 0 && green < 20) green = 20;
	if (blue != 0 && blue < 20) blue = 20;
	
	int fadeTime;
	if (red == 0 && blue == 0 && green ==0) fadeTime = display.rgbLed[rgbLedNumber].fadeOffSpeed;
	else fadeTime = display.rgbLed[rgbLedNumber].fadeOnSpeed;
	
	// Calculate the difference between the current colour and the new colour
	// this ensures the fade speed is the same no matter the difference in brightness levels
	// between the three colours
	//
	// The fade speed is defined in tenths of a second, we have 4095 brightness levels so to fade
	// in one tenth of a second (at 61.3 interrupts per second which is 6.13 interrupts per tenth of a
	// second.
	//
	// Therefore if the difference between the actual brightness and the target brightness is 
	// (target - actual) we can calculate this as:
	//
	// ((target brightness - desired brightness) / fadeTime ) / 6.13)
	//
	float redFadeSpeedFloat =
		(( (float)display.rgbLed[rgbLedNumber].redTarget - red) / (float)fadeTime) / 6.13;
		
	float greenFadeSpeedFloat =
		(( (float)display.rgbLed[rgbLedNumber].greenTarget - green) / (float)fadeTime) / 6.13;
		
	float blueFadeSpeedFloat =
		(( (float)display.rgbLed[rgbLedNumber].blueTarget - blue) / (float)fadeTime) / 6.13;
	
	// Set the fade speed as an integer value
	display.rgbLed[rgbLedNumber].redFadeSpeed = (int)redFadeSpeedFloat;
	display.rgbLed[rgbLedNumber].greenFadeSpeed = (int)greenFadeSpeedFloat;
	display.rgbLed[rgbLedNumber].blueFadeSpeed = (int)blueFadeSpeedFloat;
	
	// Ensure that the fade speed increment is always positive
	if (display.rgbLed[rgbLedNumber].redFadeSpeed < 0)
		display.rgbLed[rgbLedNumber].redFadeSpeed = -display.rgbLed[rgbLedNumber].redFadeSpeed;
	if (display.rgbLed[rgbLedNumber].greenFadeSpeed < 0)
		display.rgbLed[rgbLedNumber].greenFadeSpeed = -display.rgbLed[rgbLedNumber].greenFadeSpeed;
	if (display.rgbLed[rgbLedNumber].blueFadeSpeed < 0)
		display.rgbLed[rgbLedNumber].blueFadeSpeed = -display.rgbLed[rgbLedNumber].blueFadeSpeed;
	
	// If the fade speed is too fast we set it to -1 which tells the
	// interrupt routine to change brightness immediately
	if (display.rgbLed[rgbLedNumber].redFadeSpeed == 0) display.rgbLed[rgbLedNumber].redFadeSpeed = -1;
	if (display.rgbLed[rgbLedNumber].greenFadeSpeed == 0) display.rgbLed[rgbLedNumber].greenFadeSpeed = -1;
	if (display.rgbLed[rgbLedNumber].blueFadeSpeed == 0) display.rgbLed[rgbLedNumber].blueFadeSpeed = -1;
	
	// Set the target brightness levels
	display.rgbLed[rgbLedNumber].redTarget = red;
	display.rgbLed[rgbLedNumber].greenTarget = green;
	display.rgbLed[rgbLedNumber].blueTarget = blue;
	
	// Reset the fade counters
	display.rgbLed[rgbLedNumber].redFadeCounter = 0;
	display.rgbLed[rgbLedNumber].greenFadeCounter = 0;
	display.rgbLed[rgbLedNumber].blueFadeCounter = 0;
}	

// Set an RBG LED fade on and off speed (the unit value is tenths of a second)
void setRgbLedFadeSpeed(unsigned char rgbLedNumber, int fadeOn, int fadeOff)
{
	display.rgbLed[rgbLedNumber].fadeOnSpeed = fadeOn;
	display.rgbLed[rgbLedNumber].fadeOffSpeed = fadeOff;
}	

// Initialise the display
void initialiseRgbDisplay(void)
{
	// Get the TLC5940s ready for action
	initialiseTlc5940();
	
	// Write the default values to the display state model
	for (int rgbLedNumber = 0; rgbLedNumber < 16; rgbLedNumber++)
	{
		display.rgbLed[rgbLedNumber].redTarget = 0;
		display.rgbLed[rgbLedNumber].greenTarget = 0;
		display.rgbLed[rgbLedNumber].blueTarget = 0;
		
		display.rgbLed[rgbLedNumber].redActual = 0;
		display.rgbLed[rgbLedNumber].greenActual = 0;
		display.rgbLed[rgbLedNumber].blueActual = 0;
		
		display.rgbLed[rgbLedNumber].redFadeSpeed = 0;
		display.rgbLed[rgbLedNumber].greenFadeSpeed = 0;
		display.rgbLed[rgbLedNumber].blueFadeSpeed = 0;
		
		display.rgbLed[rgbLedNumber].redFadeCounter = 0;
		display.rgbLed[rgbLedNumber].greenFadeCounter = 0;
		display.rgbLed[rgbLedNumber].blueFadeCounter = 0;
		
		display.rgbLed[rgbLedNumber].fadeOnSpeed = 0;
		display.rgbLed[rgbLedNumber].fadeOffSpeed = 0;
	}
	
	display.brightnessCounter = 0;	
}	

#endif